from typing import Protocol

from bazario.asyncio.protocols.handle_next import HandleNext
from bazario.asyncio.protocols.resolver import Resolver
from bazario.typing.type_vars import TRes_co, TTarget_contra


class PipelineBehaviour(Protocol[TTarget_contra, TRes_co]):
    async def handle(
        self,
        resolver: Resolver,
        target: TTarget_contra,
        handle_next: HandleNext[TTarget_contra, TRes_co],
    ) -> TRes_co: ...



from bazario.protocols.pipeline_behaviour import PipelineBehaviour
from bazario.typing.type_vars import TRes_co, TTarget_contra


class PipelineBehaviourRegistry:
    def __init__(self) -> None:
        self._behaviours: dict[
            type[TTarget_contra],
            list[PipelineBehaviour[TTarget_contra, TRes_co]],
        ] = {}

    def add_behaviours(
        self,
        target_type: type[TTarget_contra],
        *behaviours: PipelineBehaviour[TTarget_contra, TRes_co],
    ) -> None:
        self._behaviours.setdefault(target_type, []).extend(behaviours)

    def get_behaviours(
        self,
        target_type: type[TTarget_contra],
    ) -> list[PipelineBehaviour[TTarget_contra, TRes_co]]:
        return [
            behaviour
            for target, behaviours in self._behaviours.items()
            if issubclass(target_type, target)
            for behaviour in behaviours
        ]



from typing import Protocol, TypeVar, runtime_checkable

from bazario.markers import Request

TRes = TypeVar("TRes")


@runtime_checkable
class Sender(Protocol):
    def send(self, request: Request[TRes]) -> TRes: ...



from typing import Protocol, runtime_checkable

from bazario.typing.type_vars import TDependency


@runtime_checkable
class Resolver(Protocol):
    async def resolve(
        self,
        dependency_type: type[TDependency],
    ) -> TDependency: ...



from typing import Protocol, TypeVar, runtime_checkable

from bazario.markers import Notification, Request

TRes_co = TypeVar("TRes_co", covariant=True)
TReq_contra = TypeVar("TReq_contra", bound=Request, contravariant=True)
TNot_contra = TypeVar("TNot_contra", bound=Notification, contravariant=True)


@runtime_checkable
class RequestHandler(Protocol[TReq_contra, TRes_co]):
    async def handle(self, request: TReq_contra) -> TRes_co: ...


@runtime_checkable
class NotificationHandler(Protocol[TNot_contra]):
    async def handle(self, notification: TNot_contra) -> None: ...



from bazario.dispatcher import Dispatcher
from bazario.markers import Notification, Request
from bazario.pipeline.behaviour_registry import PipelineBehaviourRegistry
from bazario.protocols.finder import HandlerFinder
from bazario.protocols.handle_next import HandleNext
from bazario.protocols.handler import NotificationHandler, RequestHandler
from bazario.protocols.pipeline_behaviour import PipelineBehaviour
from bazario.protocols.publisher import Publisher
from bazario.protocols.resolver import Resolver
from bazario.protocols.sender import Sender

__all__ = (
    "Dispatcher",
    "HandleNext",
    "HandlerFinder",
    "Notification",
    "NotificationHandler",
    "PipelineBehaviour",
    "PipelineBehaviourRegistry",
    "Publisher",
    "Request",
    "RequestHandler",
    "Resolver",
    "Sender",
)



from typing import TypeVar

from bazario.markers import Notification, Request

TDependency = TypeVar("TDependency")
TRes_co = TypeVar("TRes_co", covariant=True)
TReq_contra = TypeVar("TReq_contra", bound=Request, contravariant=True)
TNot_contra = TypeVar("TNot_contra", bound=Notification, contravariant=True)
TTarget_contra = TypeVar(
    "TTarget_contra",
    contravariant=True,
    bound=Request | Notification,
)



from typing import Protocol, runtime_checkable

from bazario.typing.type_vars import TNot_contra, TReq_contra, TRes_co


@runtime_checkable
class RequestHandler(Protocol[TReq_contra, TRes_co]):
    def handle(self, request: TReq_contra) -> TRes_co: ...


@runtime_checkable
class NotificationHandler(Protocol[TNot_contra]):
    def handle(self, notification: TNot_contra) -> None: ...



from collections.abc import Iterable
from typing import get_args, get_origin

ORIG_BASES = "__orig_bases__"


def extract_orig_bases(cls: type) -> Iterable:
    return getattr(cls, ORIG_BASES, ())


def extract_base_generic_type(cls: type) -> type | None:
    orig_bases = extract_orig_bases(cls)

    for orig_base in orig_bases:
        if get_origin(orig_base) is not None:
            return orig_base

    return None


def matches_generic_type(
    hint: type,
    base_generic: type,
    expected_arg: type,
) -> bool:
    """
    Checks whether the type `hint` is a generalized type
    derived from `base_generic` and whether it contains
    `expected_arg` as a type argument.
    """

    if get_origin(hint) is not base_generic:
        return False

    args = get_args(hint)

    return expected_arg in args



from bazario.asyncio.dispatcher import Dispatcher
from bazario.asyncio.pipeline.behaviour_registry import (
    PipelineBehaviourRegistry,
)
from bazario.asyncio.protocols.finder import HandlerFinder
from bazario.asyncio.protocols.handle_next import HandleNext
from bazario.asyncio.protocols.handler import (
    NotificationHandler,
    RequestHandler,
)
from bazario.asyncio.protocols.pipeline_behaviour import PipelineBehaviour
from bazario.asyncio.protocols.publisher import Publisher
from bazario.asyncio.protocols.resolver import Resolver
from bazario.asyncio.protocols.sender import Sender

__all__ = (
    "Dispatcher",
    "HandleNext",
    "HandlerFinder",
    "NotificationHandler",
    "PipelineBehaviour",
    "PipelineBehaviourRegistry",
    "Publisher",
    "RequestHandler",
    "Resolver",
    "Sender",
)



from collections.abc import Iterable
from typing import Protocol, runtime_checkable

from bazario.asyncio.protocols.handler import (
    NotificationHandler,
    RequestHandler,
)
from bazario.markers import Notification, Request


@runtime_checkable
class HandlerFinder(Protocol):
    async def find_with_request(
        self,
        request_type: type[Request],
    ) -> type[RequestHandler] | None: ...
    async def find_with_notification(
        self,
        notification_type: type[Notification],
    ) -> Iterable[type[NotificationHandler]]: ...



__all__ = (
    "DishkaHandlerFinder",
    "DishkaResolver",
)

from collections.abc import Iterable

from dishka import AsyncContainer

from bazario import Notification, Request
from bazario.asyncio import (
    NotificationHandler,
    RequestHandler,
    Resolver,
)
from bazario.asyncio.protocols.finder import HandlerFinder
from bazario.typing.type_inspection import (
    extract_base_generic_type,
    matches_generic_type,
)
from bazario.typing.type_vars import TDependency


class DishkaResolver(Resolver):
    def __init__(self, container: AsyncContainer) -> None:
        self._container = container

    async def resolve(self, dependency_type: type[TDependency]) -> TDependency:
        return await self._container.get(dependency_type)


class DishkaHandlerFinder(HandlerFinder):
    def __init__(self, container: AsyncContainer) -> None:
        self._factories = container.registry.factories

    async def find_with_request(
        self,
        request_type: type[Request],
    ) -> type[RequestHandler] | None:
        for key in self._factories:
            generic_type = extract_base_generic_type(key.type_hint)

            matches = matches_generic_type(
                generic_type,
                RequestHandler,
                request_type,
            )

            if generic_type and matches:
                return key.type_hint

        return None

    async def find_with_notification(
        self,
        notification_type: type[Notification],
    ) -> Iterable[type[NotificationHandler]]:
        handler_types = []
        for key in self._factories:
            generic_type = extract_base_generic_type(key.type_hint)

            matches = matches_generic_type(
                generic_type,
                NotificationHandler,
                notification_type,
            )

            if generic_type and matches:
                handler_types.append(key.type_hint)

        return handler_types



from bazario.exceptions import HandlerNotFoundError
from bazario.markers import Notification, Request
from bazario.pipeline.behaviour_registry import PipelineBehaviourRegistry
from bazario.pipeline.wrap_pipeline_behaviours import wrap_pipeline_behaviours
from bazario.protocols.finder import HandlerFinder
from bazario.protocols.publisher import Publisher
from bazario.protocols.resolver import Resolver
from bazario.protocols.sender import Sender, TRes


class Dispatcher(Sender, Publisher):
    def __init__(
        self,
        resolver: Resolver,
        handler_finder: HandlerFinder,
        pipeline_behaviour_registry: PipelineBehaviourRegistry,
    ) -> None:
        self._resolver = resolver
        self._handler_finder = handler_finder
        self._pipeline_behaviour_registry = pipeline_behaviour_registry

    def send(self, request: Request[TRes]) -> TRes:
        request_type = type(request)
        handler_type = self._handler_finder.find_with_request(request_type)

        if handler_type is None:
            raise HandlerNotFoundError(request_type)

        handler = self._resolver.resolve(handler_type)

        behaviours = self._pipeline_behaviour_registry.get_behaviours(
            request_type,
        )

        pipeline = wrap_pipeline_behaviours(behaviours, handler)

        return pipeline(self._resolver, request)

    def publish(self, notification: Notification) -> None:
        notification_type = type(notification)
        handler_types = self._handler_finder.find_with_notification(
            notification_type,
        )
        behaviours = self._pipeline_behaviour_registry.get_behaviours(
            notification_type,
        )

        for handler_type in handler_types:
            handler = self._resolver.resolve(handler_type)
            pipeline = wrap_pipeline_behaviours(behaviours, handler)
            pipeline(self._resolver, notification)



from typing import Protocol, runtime_checkable

from bazario.markers import Notification


@runtime_checkable
class Publisher(Protocol):
    async def publish(self, notification: Notification) -> None: ...



from typing import Protocol, runtime_checkable

from bazario.typing.type_vars import TDependency


@runtime_checkable
class Resolver(Protocol):
    def resolve(self, dependency_type: type[TDependency]) -> TDependency: ...



from functools import partial
from typing import TypeAlias

from bazario.asyncio.protocols.handle_next import HandleNext
from bazario.asyncio.protocols.handler import (
    NotificationHandler,
    RequestHandler,
)
from bazario.asyncio.protocols.pipeline_behaviour import PipelineBehaviour
from bazario.asyncio.protocols.resolver import Resolver
from bazario.typing.type_vars import TRes_co, TTarget_contra

_HandlerType: TypeAlias = (
    RequestHandler[TTarget_contra, TRes_co]
    | NotificationHandler[TTarget_contra]
)


def wrap_pipeline_behaviours(
    behaviours: list[PipelineBehaviour[TTarget_contra, TRes_co]],
    handler: _HandlerType,
) -> HandleNext[TTarget_contra, TRes_co]:
    async def process_pipeline(
        resolver: Resolver,
        target: TTarget_contra,
    ) -> TRes_co:
        async def handle_next(
            resolver: Resolver,
            target: TTarget_contra,
        ) -> TRes_co:
            return await handler.handle(target)

        for behaviour in reversed(behaviours):
            handle_next = partial(behaviour.handle, handle_next=handle_next)

        return await handle_next(resolver, target)

    return process_pipeline
